/* sort-lines - Masanao Izumo <mo@goice.co.jp> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#define MMAP_USE
/* #define PERFORMANCE_ANALYZE */
/* #define SVR4 */

char* version = "1.2";

#ifdef MMAP_USE
#include <sys/mman.h>
#endif /* MMAP_USE */

#ifdef PERFORMANCE_ANALYZE
#include <sys/time.h>
static double get_current_calender_time(void)
{
    struct timeval tv;

#ifdef SVR4
    gettimeofday(&tv);
#else
    gettimeofday(&tv, NULL);
#endif

    return (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0 ;
}
#endif /* PERFORMANCE_ANALYZE */


char* readfile(const char* name, size_t* size)
{
    int fd;
    struct stat buf;
    char* buff;

    if((fd = open(name, O_RDONLY)) < 0)
    {
	perror(name);
	return NULL;
    }

    if(fstat(fd, &buf) < 0)
    {
	perror(name);
	close(fd);
	return NULL;
    }
    *size = buf.st_size;

#ifdef MMAP_USE
    if((buff = mmap(0, *size, PROT_READ, MAP_PRIVATE, fd, 0)) == (caddr_t) - 1)
    {
	perror(name);
	close(fd);
	return NULL;
    }
#else
    if((buff = (char *)malloc(*size + 1)) == NULL)
    {
	perror("malloc");
	exit(1);
    }
    read(fd, buff, *size);
    buff[*size] = '\0';
#endif /* MMAP_USE */

    close(fd);

    return buff;
}


/* memory utils
 * to decrease malloc function
 */
char* alloc_chars(long n)
{
    static char* buffer;
    static int rest;
    char* p;

    if(n >= 1024)
    {
	if((p = (char *)malloc(n)) == NULL)
	{
	    perror("malloc");
	    exit(1);
	}
	return p;
    }

    if(buffer == NULL || rest - n < 0)
    {
	rest = 1024*1024;
	if((buffer = (char *)malloc(rest)) == NULL)
	{
	    perror("malloc");
	    exit(1);
	}
    }

    p = buffer;
    buffer += n;
    rest -= n;

    return p;
}
/* end of memory utils */


#define BLOCKSIZE 65536
struct string_list
{
    int n;
    char* string[BLOCKSIZE];
    struct string_list* next;
};

struct string_list* string_list = NULL;
static long string_list_len = 0;
unsigned char** line_array;

void make_line_array(void)
{
    long i, j, n;
    struct string_list* node;

    if((line_array = (unsigned char **)malloc(string_list_len *
					      sizeof(unsigned char *))) == NULL)
    {
	perror("malloc");
	exit(1);
    }

    for(i = 0, node = string_list; node; node = node->next)
    {
	n = node->n;
	for(j = 0; j < n; j++, i++)
	    line_array[i] = (unsigned char *)node->string[j];
    }
}

void add_line(const char* line, long len)
{
    char* s;

    if(len > 0 && line[len - 1] == '\n')
    {
	len--;
        if(len > 0 && line[len - 1] == '\r')
	    len--;
    }

    if(string_list == NULL || string_list->n == BLOCKSIZE)
    {
	struct string_list* node;

	if((node = (struct string_list *)malloc(sizeof(struct string_list))) ==
	   NULL)
	{
	    perror("malloc");
	    exit(1);
	}
	node->n = 0;
	node->next = string_list;
	string_list = node;
    }

    s = alloc_chars(len + 1);
    memcpy(s, line, len);
    s[len] = '\0';
    string_list->string[string_list->n++] = s;
    string_list_len++;
}

void read_lines(const char* filename)
{
    if(!strcmp(filename, "-"))	/* read from stdin */
    {
	char buff[BUFSIZ];
	while(fgets(buff, BUFSIZ, stdin) != NULL)
	    add_line(buff, strlen(buff));
	rewind(stdin);
    }
    else
    {
	char* file;
	size_t n;
	char* p;
	char* line;
	long len;

	file = readfile(filename, &n);

	if(file == NULL)
	    return;
	p = file;
	while(1)
	{
	    line = p;
	    p = strchr(p, '\n');
	    if(p != NULL)
	    {
		p++;
		len = p - line;
		add_line(line, len);
	    }
	    else
	    {
		char* q;
		for(q = line, len = 0; *q; q++, len++)
		    ;
		if(len > 0)
		    add_line(line, len);
		break;
	    }
	}
#ifdef MMAP_USE
	munmap(file, n);
#else
	free(file);
#endif /* MMAP_USE */
    }
}

/*
 * fast sorting function
 */
/* LESSP(x, y) means x is less than y */
#define LESSP(x, y) (strcmp((x), (y)) < 0)
typedef const unsigned char* qsort_datatype;
#define SORT_THRESHOLD 20

#ifdef SORT_THRESHOLD
static void my_inssort(qsort_datatype* data, long n)
{
    long i, j;
    qsort_datatype x;

    for(i = 1; i < n; i++)
    {
	x = data[i];
	for(j = i - 1; j >= 0 && LESSP(x, data[j]); j--)
	    data[j + 1] = data[j];
	data[j + 1] = x;
    }
}
#endif /* SORT_THRESHOLD */

static void my_qsort(qsort_datatype* a, long first, long last)
{
    long i = first, j = last;
    qsort_datatype x, t;

#ifdef SORT_THRESHOLD
    if(j - i < SORT_THRESHOLD)
    {
	my_inssort(a + i, j - i + 1);
	return;
    }
#endif /* SORT_THRESHOLD */

    x = a[(first + last) / 2];

    for(;;)
    {
	while(LESSP(a[i], x))
	    i++;
	while(LESSP(x, a[j]))
	    j--;
	if(i >= j)
	    break;
	t = a[i];
	a[i] = a[j];
	a[j] = t;
	i++;
	j--;
    }
    if(first < i - 1)
	my_qsort(a, first, i - 1);
    if(j + 1 < last)
	my_qsort(a, j + 1, last);
}

void sort_lines(void)
{
    my_qsort((qsort_datatype *)line_array, 0, string_list_len - 1);
}

int uniq_flag = 0;
void print_lines(void)
{
    int i;

    if(!uniq_flag)
	for(i = 0; i < string_list_len; i++)
	    puts(line_array[i]);
    else
    {
	unsigned char* prevline;

	if(string_list_len == 0)
	    return;
	puts(line_array[0]);
	prevline = line_array[0];
	for(i = 1; i < string_list_len; i++)
	    if(strcmp(prevline, line_array[i]) != 0)
	    {
		puts(line_array[i]);
		prevline = line_array[i];
	    }
    }
}

char* progname;
void usage(void)
{
    fprintf(stderr, "Usage: %s [-h] [-V] [-u] [--] [files ...]\n", progname);
}

int main(int argc, char** argv)
{
    int i, argstart = 1;
#ifdef PERFORMANCE_ANALYZE
    double t;
#endif /* PERFORMANCE_ANALYZE */

    if((progname = strrchr(argv[0], '/')) == NULL)
	progname = argv[0];
    else
	progname++;

    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-h"))
	{
	    usage();
	    return 0;
	}
	else if(!strcmp(argv[i], "-V"))
	{
	    printf("%s version %s\n", progname, version);
	    return 0;
	}
	else if(!strcmp(argv[i], "-u"))
	{
	    uniq_flag = 1;
	}
	else if(!strcmp(argv[i], "--"))
	{
	    i++;
	    break;
	}
	else if(argv[i][0] == '-')
	{
	    usage();
	    return 1;
	}
	else
	    break;
    }
    argstart = i;

#ifdef PERFORMANCE_ANALYZE
    t = get_current_calender_time();
    fprintf(stderr, "Reading lines...");
#endif /* PERFORMANCE_ANALYZE */

    if(argc == argstart)
	read_lines("-");
    else
	for(i = argstart; i < argc; i++)
	    read_lines(argv[i]);

#ifdef PERFORMANCE_ANALYZE
    fprintf(stderr, "Done (%g sec.)\n", get_current_calender_time() - t);
    t = get_current_calender_time();
    fprintf(stderr, "Making line array...");
#endif /* PERFORMANCE_ANALYZE */

    make_line_array();

#ifdef PERFORMANCE_ANALYZE
    fprintf(stderr, "Done (%g sec.)\n", get_current_calender_time() - t);
    t = get_current_calender_time();
    fprintf(stderr, "Sorting lines...");
#endif /* PERFORMANCE_ANALYZE */

    sort_lines();

#ifdef PERFORMANCE_ANALYZE
    fprintf(stderr, "Done (%g sec.)\n", get_current_calender_time() - t);
#endif /* PERFORMANCE_ANALYZE */

    print_lines();
    return 0;
}
